@@ -1,25 +1,25 @@
 /*
- *  Yaboot - secondary boot loader for Linux on PowerPC. 
+ *  Yaboot - secondary boot loader for Linux on PowerPC.
  *
- *  Copyright (C) 2001 Ethan Benson
+ *  Copyright (C) 2001, 2002 Ethan Benson
  *
  *  Copyright (C) 1999, 2000, 2001 Benjamin Herrenschmidt
- *  
+ *
  *  IBM CHRP support
- *  
+ *
  *  Copyright (C) 2001 Peter Bergner
  *
  *  portions based on poof
- *  
+ *
  *  Copyright (C) 1999 Marius Vollmer
- *  
+ *
  *  portions based on quik
- *  
+ *
  *  Copyright (C) 1996 Paul Mackerras.
- *  
+ *
  *  Because this program is derived from the corresponding file in the
  *  silo-0.64 distribution, it is also
- *  
+ *
  *  Copyright (C) 1996 Pete A. Zaitcev
  *                1996 Maurizio Plaza
  *                1996 David S. Miller
@@ -111,7 +111,9 @@ static void     setup_display(void);
 /* Locals & globals */
 
 int useconf = 0;
-char bootdevice[1024];
+char bootdevice[BOOTDEVSZ];
+char bootoncelabel[1024];
+char bootargs[1024];
 char *password = NULL;
 struct boot_fspec_t boot;
 int _machine = _MACH_Pmac;
@@ -180,7 +182,7 @@ yaboot_start (unsigned long r3, unsigned long r4, unsigned long r5)
 
      /* OF seems to do it, but I'm not very confident */
      memset(&__bss_start, 0, &_end - &__bss_start);
-  	
+
      /* Check for quik first stage bootloader (but I don't think we are
       * compatible with it anyway, I'll look into backporting to older OF
       * versions later
@@ -192,7 +194,7 @@ yaboot_start (unsigned long r3, unsigned long r4, unsigned long r5)
 
      /* Initialize OF interface */
      prom_init ((prom_entry) r5);
-	
+
      /* Allocate some memory for malloc'ator */
      malloc_base = prom_claim((void *)MALLOCADDR, MALLOCSIZE, 0);
      if (malloc_base == (void *)-1) {
@@ -203,7 +205,7 @@ yaboot_start (unsigned long r3, unsigned long r4, unsigned long r5)
      malloc_init(malloc_base, MALLOCSIZE);
      DEBUG_F("Malloc buffer allocated at %p (%d bytes)\n",
 	     malloc_base, MALLOCSIZE);
-		
+
      /* A few useless DEBUG_F's */
      DEBUG_F("reloc_offset :  %ld         (should be 0)\n", reloc_offset());
      DEBUG_F("test_bss     :  %d         (should be 0)\n", test_bss);
@@ -226,7 +228,7 @@ yaboot_start (unsigned long r3, unsigned long r4, unsigned long r5)
 		    _machine = _MACH_chrp;
 	  }
      }
-	
+
      DEBUG_F("Running on _machine = %d\n", _machine);
      DEBUG_SLEEP;
 
@@ -241,9 +243,9 @@ yaboot_start (unsigned long r3, unsigned long r4, unsigned long r5)
 
      /* Return to OF */
      prom_exit();
-	
+
      return result;
-	
+
 }
 
 #ifdef CONFIG_COLOR_TEXT
@@ -260,13 +262,13 @@ check_color_text_ui(char *color)
 	  i++;
      }
      return -1;
-}      
+}
 #endif /* CONFIG_COLOR_TEXT */
 
 
 void print_message_file(char *filename)
 {
-     char *msg = NULL; 
+     char *msg = NULL;
      char *p, *endp;
      char *defdev = boot.dev;
      int defpart = boot.part;
@@ -284,7 +286,7 @@ void print_message_file(char *filename)
 	  if (p) {
 	       n = simple_strtol(p, &endp, 10);
 	       if (endp != p && *endp == 0)
-		    defpart = n;	       
+		    defpart = n;
 	  }
 
      strncpy(msgpath, filename, sizeof(msgpath));
@@ -324,13 +326,11 @@ done:
  * config file. Handle the "\\" (blessed system folder)
  */
 static int
-load_config_file(char *device, char* path, int partition)
+load_config_file(struct boot_fspec_t *fspec)
 {
      char *conf_file = NULL, *p;
      struct boot_file_t file;
      int sz, opened = 0, result = 0;
-     char conf_path[512];
-     struct boot_fspec_t fspec;
 
      /* Allocate a buffer for the config file */
      conf_file = malloc(CONFIG_FILE_MAX);
@@ -339,23 +339,11 @@ load_config_file(char *device, char* path, int partition)
 	  goto bail;
      }
 
-     /* Build the path to the file */
-     if (_machine == _MACH_chrp)
-	  strcpy(conf_path, "/etc/");
-     else if (path && *path)
-	  strcpy(conf_path, path);
-     else
-	  conf_path[0] = 0;
-     strcat(conf_path, CONFIG_FILE_NAME);
-
      /* Open it */
-     fspec.dev = device;
-     fspec.file = conf_path;
-     fspec.part = partition;
-     result = open_file(&fspec, &file);
+     result = open_file(fspec, &file);
      if (result != FILE_ERR_OK) {
-	  prom_printf("%s:%d,", fspec.dev, fspec.part);
-	  prom_perror(result, fspec.file);
+	  prom_printf("%s:%d,", fspec->dev, fspec->part);
+	  prom_perror(result, fspec->file);
 	  prom_printf("Can't open config file\n");
 	  goto bail;
      }
@@ -375,11 +363,27 @@ load_config_file(char *device, char* path, int partition)
      opened = 0;
 
      /* Call the parsing code in cfg.c */
-     if (cfg_parse(conf_path, conf_file, sz) < 0) {
+     if (cfg_parse(fspec->file, conf_file, sz) < 0) {
 	  prom_printf ("Syntax error or read error config\n");
 	  goto bail;
      }
 
+     /* 
+      * set the default cf_option to label that has the same MAC addr 
+      * it only works if there is a label with the MAC addr on yaboot.conf
+      */
+     if (prom_get_devtype(fspec->dev) == FILE_DEVICE_NET) {
+         /* change the variable bellow to get the MAC dinamicaly */
+         char * macaddr = NULL;
+         int default_mac = 0;
+
+         macaddr = prom_get_mac(prom_get_netinfo());
+         default_mac = cfg_set_default_by_mac(macaddr);
+         if (default_mac >= 1) {
+            prom_printf("Default label was changed to macaddr label.\n");
+         }
+     }
+
      DEBUG_F("Config file successfully parsed, %d bytes\n", sz);
 
      /* Now, we do the initialisations stored in the config file */
@@ -388,7 +392,7 @@ load_config_file(char *device, char* path, int partition)
 	  prom_interpret(p);
 
      password = cfg_get_strg(0, "password");
-	
+
 #ifdef CONFIG_COLOR_TEXT
      p = cfg_get_strg(0, "fgcolor");
      if (p) {
@@ -407,19 +411,19 @@ load_config_file(char *device, char* path, int partition)
      }
      if (bgcolor >= 0) {
 	  char temp[64];
-	  sprintf(temp, "%x to background-color", bgcolor); 
-	  prom_interpret(temp); 
+	  sprintf(temp, "%x to background-color", bgcolor);
+	  prom_interpret(temp);
 #if !DEBUG
 	  prom_printf("\xc");
 #endif /* !DEBUG */
      }
      if (fgcolor >= 0) {
 	  char temp[64];
-	  sprintf(temp, "%x to foreground-color", fgcolor); 
-	  prom_interpret(temp); 
+	  sprintf(temp, "%x to foreground-color", fgcolor);
+	  prom_interpret(temp);
      }
 #endif /* CONFIG_COLOR_TEXT */
-   
+
      p = cfg_get_strg(0, "init-message");
      if (p)
 	  prom_printf("%s\n", p);
@@ -429,18 +433,74 @@ load_config_file(char *device, char* path, int partition)
 	  print_message_file(p);
 
      result = 1;
-    
+
 bail:
 
      if (opened)
 	  file.fs->close(&file);
-    
-     if (result != 1 && conf_file)
+
+     if (conf_file)
 	  free(conf_file);
-    	
+
      return result;
 }
 
+/*
+ * Search for config file by MAC address, then by IP address.
+ * Basically copying pxelinux's algorithm.
+ * http://syslinux.zytor.com/pxe.php#config
+ */
+static int load_my_config_file(struct boot_fspec_t *orig_fspec)
+{
+     struct bootp_packet *packet;
+     int rc = 0;
+     struct boot_fspec_t fspec = *orig_fspec;
+     char *cfgpath = (_machine == _MACH_chrp) ? "/etc/" : "";
+     int flen;
+     int minlen;
+
+     packet = prom_get_netinfo();
+
+     /*
+      * First, try to match on mac address with the hardware type
+      * prepended.
+      */
+
+     /* 3 chars per byte in chaddr + 2 chars for htype + /etc/ +\0 */
+     fspec.file = malloc(packet->hlen * 3 + 2 + 6);
+     if (!fspec.file)
+	  goto out;
+
+     sprintf(fspec.file, "%s%02x-", cfgpath, packet->htype);
+     strcat(fspec.file, prom_get_mac(packet));
+
+     rc = load_config_file(&fspec);
+     if (rc)
+	  goto out;
+
+     /*
+      * Now try to match on IP.
+      */
+     /* no need to realloc for /etc/ + 8 chars in yiaddr + \0 */
+     sprintf(fspec.file, "%s%s", cfgpath, prom_get_ip(packet));
+
+     for (flen = strlen(fspec.file),
+          minlen = strlen(cfgpath); flen > minlen; flen--) {
+	  rc = load_config_file(&fspec);
+	  if (rc)
+	       goto out;
+	  /* Chop one digit off the end, try again */
+	  fspec.file[flen - 1] = '\0';
+     }
+
+ out:
+     if (rc) /* modify original only on success */
+	  orig_fspec->file = fspec.file;
+     else
+	  free(fspec.file);
+     return rc;
+}
+
 void maintabfunc (void)
 {
      if (useconf) {
@@ -559,7 +619,7 @@ void check_password(char *str)
 	  if (!strncmp (password, "$1$", 3)) {
 	       if (!check_md5_password(passwdbuff, password))
 		    return;
-	  } 
+	  }
 	  else if (!strcmp (password, passwdbuff))
 	       return;
 #else /* !MD5 */
@@ -582,6 +642,7 @@ int get_params(struct boot_param_t* params)
 {
      int defpart;
      char *defdevice = 0;
+     char defdevice_bak[1024];
      char *p, *q, *endp;
      int c, n;
      char *imagename = 0, *label;
@@ -590,7 +651,6 @@ int get_params(struct boot_param_t* params)
      int singlekey = 0;
      int restricted = 0;
      static int first = 1;
-     static char bootargs[1024];
      static char imagepath[1024];
      static char initrdpath[1024];
      static char sysmappath[1024];
@@ -602,12 +662,11 @@ int get_params(struct boot_param_t* params)
      params->rd.part = -1;
      params->sysmap.part = -1;
      defpart = boot.part;
-    
+
      cmdinit();
 
      if (first) {
 	  first = 0;
-	  prom_get_chosen("bootargs", bootargs, sizeof(bootargs));
 	  imagename = bootargs;
 	  word_split(&imagename, &params->args);
 	  timeout = DEFAULT_TIMEOUT;
@@ -654,8 +713,12 @@ int get_params(struct boot_param_t* params)
      }
 
      if (c == '\n' || c == '\r') {
-	  if (!imagename)
-	       imagename = cfg_get_default();
+	  if (!imagename) {
+	       if (bootoncelabel[0] != 0)
+		    imagename = bootoncelabel;
+	       else
+		    imagename = cfg_get_default();
+	  }
 	  if (imagename)
 	       prom_printf("%s", imagename);
 	  if (params->args)
@@ -677,6 +740,8 @@ int get_params(struct boot_param_t* params)
      label = 0;
      defdevice = boot.dev;
 
+     strcpy(defdevice_bak,defdevice);
+
      if (useconf) {
 	  defdevice = cfg_get_strg(0, "device");
 	  p = cfg_get_strg(0, "partition");
@@ -729,10 +794,15 @@ int get_params(struct boot_param_t* params)
 	       "resides on, and \"partno\" is the partition number the image resides on.\n"
 	       "Note that the comma (,) is only required if you specify an OpenFirmware\n"
 	       "device, if you only specify a filename you should not start it with a \",\"\n\n"
-	       "If you omit \"device:\" and \"partno\" yaboot will use the values of \n"
-	       "\"device=\" and \"partition=\" in yaboot.conf, right now those are set to: \n"
-	       "device=%s\n"
-	       "partition=%d\n\n", defdevice, defpart);
+	       "To load an alternative config file rather than /etc/yaboot.conf, enter\n"
+	       "its device, partno and path, on Open Firmware Prompt:\n"
+	       "boot conf=device:partno,/path/to/configfile\n."
+	       "To reload the config file or load a new one, use the \"conf\" command\n"
+	       "on Yaboot's prompt:\n"
+	       "conf [device=device] [partition=partno] [file=/path/to/configfile]\n\n"
+	       "If you omit \"device\" and \"partno\", Yaboot will use their current\n"
+	       "values. You can check them by entering \"conf\" on Yaboot's prompt.\n");
+
 	  return 0;
      }
 
@@ -747,7 +817,92 @@ int get_params(struct boot_param_t* params)
 	       check_password ("Restricted command.");
 	       return 1;
 	  }
-	  return 1; 
+	  return 1;
+     }
+
+     if (!strncmp (imagename, "conf", 4)) {
+
+         // imagename = "conf file=blah dev=bleh part=blih"
+         DEBUG_F("Loading user-specified config file: %s\n",imagename);
+         if (password) {
+             check_password ("Restricted command.");
+             return 1;
+         }
+
+         // args= "file=blah dev=bleh part=blih"
+         char *args = params->args;
+
+         if (strlen(args)){
+
+            // set a pointer to the first space in args
+            char *space = strchr(args,' ');
+
+            int loop = 3;
+            while (loop > 0){
+                char temp[1024] = "0";
+
+                // copy next argument to temp
+                strncpy(temp, args, space-args);
+
+                // parse temp and set boot arguments
+                if (!strncmp (temp, "file=", 5)){
+                   DEBUG_F("conf file: %s\n", temp+5);
+                   strcpy(boot.file, temp+5);
+                } else if (!strncmp (temp, "device=", 7)){
+                   DEBUG_F("conf device: %s\n", temp+7);
+                   strcpy(boot.dev, temp+7);
+                } else if (!strncmp (temp, "partition=", 10)){
+                   DEBUG_F("conf partition: %s\n", temp+10);
+                   boot.part=simple_strtol(temp+10,NULL,10);
+                } else
+                   space = NULL;
+
+                // set the pointer to the next space in args;
+                // set the loop control variable
+                if (strlen(space)>1){
+                    // Go to the next argument
+                    args = space+1;
+
+                    loop--;
+                    if (strchr(args,' ') == NULL)
+                        space = &args[strlen(args)];
+                    else
+                        space = strchr(args,' ');
+                } else {
+                    loop = -1;
+                    space = NULL;
+                }
+            }
+
+            prom_printf("Loading config file...\n");
+            useconf = load_config_file(&boot);
+            if (useconf > 0){
+                if ((q = cfg_get_strg(0, "timeout")) != 0 && *q != 0)
+                   timeout = simple_strtol(q, NULL, 0);
+            } else {
+               prom_printf("Restoring default values.\n");
+               strcpy(boot.file,"");
+               strcpy(boot.dev, defdevice_bak);
+               boot.part = defpart;
+            }
+
+         } else {
+             prom_printf("Current configuration:\n");
+             prom_printf("device: %s\n", boot.dev);
+             if (boot.part < 0)
+                 prom_printf("partition: auto\n");
+             else
+                 prom_printf("partition: %d\n", boot.part);
+             if (strlen(boot.file))
+                 prom_printf("file: %s\n", boot.file);
+             else
+                 prom_printf("file: /etc/%s\n",CONFIG_FILE_NAME);
+         }
+
+         imagename = "\0";
+         params->args = "\0";
+
+         return 0;
      }
 
      if (imagename[0] == '$') {
@@ -819,7 +974,7 @@ yaboot_text_ui(void)
      loadinfo_t          loadinfo;
      void                *initrd_more,*initrd_want;
      unsigned long       initrd_read;
-    
+
      loadinfo.load_loc = 0;
 
      for (;;) {
@@ -827,12 +982,12 @@ yaboot_text_ui(void)
 	  initrd_base = 0;
 	  sysmap_base = 0;
 	  sysmap_size = 0;
-    	
+
 	  if (get_params(&params))
 	       return;
 	  if (!params.kernel.file)
 	       continue;
-	
+
 	  prom_printf("Please wait, loading kernel...\n");
 
 	  memset(&file, 0, sizeof(file));
@@ -906,7 +1061,7 @@ yaboot_text_ui(void)
 		    free(params.sysmap.file);
 		    params.sysmap.file=loc;
 	       }
-	     
+
 	       result = open_file(&params.sysmap, &file);
 	       if (result != FILE_ERR_OK) {
 		    prom_printf("%s:%d,", params.sysmap.dev, params.sysmap.part);
@@ -916,6 +1071,7 @@ yaboot_text_ui(void)
 		    sysmap_base = prom_claim(loadinfo.base+loadinfo.memsize, 0x100000, 0);
 		    if (sysmap_base == (void *)-1) {
 			 prom_printf("Claim failed for sysmap memory\n");
+			 prom_pause();
 			 sysmap_base = 0;
 		    } else {
 			 sysmap_size = file.fs->read(&file, 0xfffff, sysmap_base);
@@ -963,7 +1119,7 @@ yaboot_text_ui(void)
 		    prom_perror(result, params.rd.file);
 	       }
 	       else {
-#define INITRD_CHUNKSIZE 0x400000
+#define INITRD_CHUNKSIZE 0x100000
 		    initrd_base = prom_claim(loadinfo.base+loadinfo.memsize, INITRD_CHUNKSIZE, 0);
 		    if (initrd_base == (void *)-1) {
 			 prom_printf("Claim failed for initrd memory\n");
@@ -979,6 +1135,7 @@ yaboot_text_ui(void)
 			      initrd_more = prom_claim(initrd_want, INITRD_CHUNKSIZE, 0);
 			      if (initrd_more != initrd_want) {
 				   prom_printf("Claim failed for initrd memory at %p rc=%p\n",initrd_want,initrd_more);
+				   prom_pause();
 				   break;
 			      }
 			      initrd_read = file.fs->read(&file, INITRD_CHUNKSIZE, initrd_more);
@@ -1005,7 +1162,7 @@ yaboot_text_ui(void)
 	  DEBUG_F(" done\n");
 
 	  if (flat_vmlinux) {
-	       /* 
+	       /*
 	        * Fill new boot infos (only if booting a vmlinux).
 	        *
 	        * The birec is low on memory, probably inside the malloc pool,
@@ -1025,12 +1182,12 @@ yaboot_text_ui(void)
 	       birec->tag = BI_FIRST;
 	       birec->size = sizeof(struct bi_record);
 	       birec = (struct bi_record *)((ulong)birec + birec->size);
-	
+
 	       birec->tag = BI_BOOTLOADER_ID;
 	       sprintf( (char *)birec->data, "yaboot");
 	       birec->size = sizeof(struct bi_record) + strlen("yaboot") + 1;
 	       birec = (struct bi_record *)((ulong)birec + birec->size);
-	
+
 	       birec->tag = BI_MACHTYPE;
 	       birec->data[0] = _machine;
 	       birec->size = sizeof(struct bi_record) + sizeof(ulong);
@@ -1081,7 +1238,7 @@ load_elf32(struct boot_file_t *file, loadinfo_t *loadinfo)
      /* Read the rest of the Elf header... */
      if ((*(file->fs->read))(file, size, &e->e_version) < size) {
 	  prom_printf("\nCan't read Elf32 image header\n");
-	  return 0;
+	  goto bail;
      }
 
      DEBUG_F("Elf32 header:\n");
@@ -1100,24 +1257,24 @@ load_elf32(struct boot_file_t *file, loadinfo_t *loadinfo)
 
      if (e->e_phnum > MAX_HEADERS) {
 	  prom_printf ("Can only load kernels with one program header\n");
-	  return 0;
+	  goto bail;
      }
 
      ph = (Elf32_Phdr *)malloc(sizeof(Elf32_Phdr) * e->e_phnum);
      if (!ph) {
 	  prom_printf ("Malloc error\n");
-	  return 0;
+	  goto bail;
      }
 
      /* Now, we read the section header */
      if ((*(file->fs->seek))(file, e->e_phoff) != FILE_ERR_OK) {
 	  prom_printf ("seek error\n");
-	  return 0;
+	  goto bail;
      }
      if ((*(file->fs->read))(file, sizeof(Elf32_Phdr) * e->e_phnum, ph) !=
 	 sizeof(Elf32_Phdr) * e->e_phnum) {
 	  prom_printf ("read error\n");
-	  return 0;
+	  goto bail;
      }
 
      /* Scan through the program header
@@ -1144,7 +1301,7 @@ load_elf32(struct boot_file_t *file, loadinfo_t *loadinfo)
 
      if (loadinfo->memsize == 0) {
 	  prom_printf("Can't find a loadable segment !\n");
-	  return 0;
+	  goto bail;
      }
 
      /* leave some room (1Mb) for boot infos */
@@ -1161,7 +1318,7 @@ load_elf32(struct boot_file_t *file, loadinfo_t *loadinfo)
           loadaddr = KERNELADDR;
      } else {
           flat_vmlinux = 0;
-          loadaddr = e->e_entry;
+          loadaddr = loadinfo->load_loc;
      }
 
      /* On some systems, loadaddr may already be claimed, so try some
@@ -1173,8 +1330,8 @@ load_elf32(struct boot_file_t *file, loadinfo_t *loadinfo)
      }
      if (loadinfo->base == (void *)-1) {
 	  prom_printf("Claim error, can't allocate kernel memory\n");
-	  return 0;
-     }	
+	  goto bail;
+     }
 
      DEBUG_F("After ELF parsing, load base: %p, mem_sz: 0x%08lx\n",
 	     loadinfo->base, loadinfo->memsize);
@@ -1192,20 +1349,25 @@ load_elf32(struct boot_file_t *file, loadinfo_t *loadinfo)
 	  if ((*(file->fs->seek))(file, p->p_offset) != FILE_ERR_OK) {
 	       prom_printf ("Seek error\n");
 	       prom_release(loadinfo->base, loadinfo->memsize);
-	       return 0;
+	       goto bail;
 	  }
 	  offset = p->p_vaddr - loadinfo->load_loc;
 	  if ((*(file->fs->read))(file, p->p_filesz, loadinfo->base+offset) != p->p_filesz) {
 	       prom_printf ("Read failed\n");
 	       prom_release(loadinfo->base, loadinfo->memsize);
-	       return 0;
+	       goto bail;
 	  }
      }
 
      free(ph);
-    
+
      /* Return success at loading the Elf32 kernel */
      return 1;
+
+bail:
+     if (ph)
+       free(ph);
+     return 0;
 }
 
 static int
@@ -1220,7 +1382,7 @@ load_elf64(struct boot_file_t *file, loadinfo_t *loadinfo)
      /* Read the rest of the Elf header... */
      if ((*(file->fs->read))(file, size, &e->e_version) < size) {
 	  prom_printf("\nCan't read Elf64 image header\n");
-	  return 0;
+	  goto bail;
      }
 
      DEBUG_F("Elf64 header:\n");
@@ -1239,24 +1401,24 @@ load_elf64(struct boot_file_t *file, loadinfo_t *loadinfo)
 
      if (e->e_phnum > MAX_HEADERS) {
 	  prom_printf ("Can only load kernels with one program header\n");
-	  return 0;
+	  goto bail;
      }
 
      ph = (Elf64_Phdr *)malloc(sizeof(Elf64_Phdr) * e->e_phnum);
      if (!ph) {
 	  prom_printf ("Malloc error\n");
-	  return 0;
+	  goto bail;
      }
 
      /* Now, we read the section header */
      if ((*(file->fs->seek))(file, e->e_phoff) != FILE_ERR_OK) {
 	  prom_printf ("Seek error\n");
-	  return 0;
+	  goto bail;
      }
      if ((*(file->fs->read))(file, sizeof(Elf64_Phdr) * e->e_phnum, ph) !=
 	 sizeof(Elf64_Phdr) * e->e_phnum) {
 	  prom_printf ("Read error\n");
-	  return 0;
+	  goto bail;
      }
 
      /* Scan through the program header
@@ -1283,7 +1445,7 @@ load_elf64(struct boot_file_t *file, loadinfo_t *loadinfo)
 
      if (loadinfo->memsize == 0) {
 	  prom_printf("Can't find a loadable segment !\n");
-	  return 0;
+	  goto bail;
      }
 
      /* leave some room (1Mb) for boot infos */
@@ -1312,8 +1474,8 @@ load_elf64(struct boot_file_t *file, loadinfo_t *loadinfo)
      }
      if (loadinfo->base == (void *)-1) {
 	  prom_printf("Claim error, can't allocate kernel memory\n");
-	  return 0;
-     }	
+	  goto bail;
+     }
 
      DEBUG_F("After ELF parsing, load base: %p, mem_sz: 0x%08lx\n",
 	     loadinfo->base, loadinfo->memsize);
@@ -1331,20 +1493,25 @@ load_elf64(struct boot_file_t *file, loadinfo_t *loadinfo)
 	  if ((*(file->fs->seek))(file, p->p_offset) != FILE_ERR_OK) {
 	       prom_printf ("Seek error\n");
 	       prom_release(loadinfo->base, loadinfo->memsize);
-	       return 0;
+	       goto bail;
 	  }
 	  offset = p->p_vaddr - loadinfo->load_loc;
 	  if ((*(file->fs->read))(file, p->p_filesz, loadinfo->base+offset) != p->p_filesz) {
 	       prom_printf ("Read failed\n");
 	       prom_release(loadinfo->base, loadinfo->memsize);
-	       return 0;
+	       goto bail;
 	  }
      }
 
      free(ph);
-    
+
      /* Return success at loading the Elf64 kernel */
      return 1;
+
+bail:
+     if (ph)
+       free(ph);
+     return 0;
 }
 
 static int
@@ -1421,16 +1588,16 @@ setup_display(void)
 	       DEBUG_F("Open screen result: %p\n", scrn);
 	  }
      }
-  	
+
      if (scrn == PROM_INVALID_HANDLE) {
-	  prom_printf("No screen device found !/n");
+	  prom_printf("No screen device found !\n");
 	  return;
      }
      for(i=0;i<16;i++) {
 	  prom_set_color(scrn, i, default_colors[i*3],
 			 default_colors[i*3+1], default_colors[i*3+2]);
      }
-     prom_printf("\x1b[1;37m\x1b[2;40m");	
+     prom_printf("\x1b[1;37m\x1b[2;40m");
 #ifdef COLOR_TEST
      for (i=0;i<16; i++) {
 	  prom_printf("\x1b[%d;%dm\x1b[1;47m%s \x1b[2;40m %s\n",
@@ -1444,7 +1611,7 @@ setup_display(void)
 		      ansi_color_table[i].name,
 		      ansi_color_table[i].name);
      }
-     prom_printf("\x1b[1;37m\x1b[2;40m");	
+     prom_printf("\x1b[1;37m\x1b[2;40m");
 #endif /* COLOR_TEST */
 
 #if !DEBUG
@@ -1458,14 +1625,40 @@ int
 yaboot_main(void)
 {
      char *ptype;
+     int conf_given = 0;
+     char conf_path[1024];
 
      if (_machine == _MACH_Pmac)
 	  setup_display();
-	
-     prom_get_chosen("bootpath", bootdevice, sizeof(bootdevice));
+
+     prom_get_chosen("bootargs", bootargs, sizeof(bootargs));
+     DEBUG_F("/chosen/bootargs = %s\n", bootargs);
+     prom_get_chosen("bootpath", bootdevice, BOOTDEVSZ);
      DEBUG_F("/chosen/bootpath = %s\n", bootdevice);
+
+     /* If conf= specified on command line, it overrides
+        Usage: conf=device:partition,/path/to/conffile
+        Example: On Open Firmware Prompt, type
+                 boot conf=/pci@8000000f8000000/pci@1/pci1014,028C@1/scsi@0/sd@1,0:3,/etc/yaboot.conf */
+
+     if (!strncmp(bootargs, "conf=", 5)) {
+        DEBUG_F("Using conf argument in Open Firmware\n");
+        char *end = strchr(bootargs,' ');
+        if (end)
+            *end = 0;
+
+        strcpy(bootdevice, bootargs + 5);
+        conf_given = 1;
+        DEBUG_F("Using conf=%s\n", bootdevice);
+
+        /* Remove conf=xxx from bootargs */
+        if (end)
+            memmove(bootargs, end+1, strlen(end+1)+1);
+        else
+            bootargs[0] = 0;
+     }
      if (bootdevice[0] == 0) {
-	  prom_get_options("boot-device", bootdevice, sizeof(bootdevice));
+	  prom_get_options("boot-device", bootdevice, BOOTDEVSZ);
 	  DEBUG_F("boot-device = %s\n", bootdevice);
      }
      if (bootdevice[0] == 0) {
@@ -1473,6 +1666,14 @@ yaboot_main(void)
 	  return -1;
      }
 
+     if (bootoncelabel[0] == 0) {
+	  prom_get_options("boot-once", bootoncelabel, 
+			   sizeof(bootoncelabel));
+     	  if (bootoncelabel[0] != 0)
+		DEBUG_F("boot-once: [%s]\n", bootoncelabel);
+     }
+     prom_set_options("boot-once", NULL, 0);
+
      if (!parse_device_path(bootdevice, NULL, -1, "", &boot)) {
 	  prom_printf("%s: Unable to parse\n", bootdevice);
 	  return -1;
@@ -1480,29 +1681,46 @@ yaboot_main(void)
      DEBUG_F("After parse_device_path: dev=%s, part=%d, file=%s\n",
 	     boot.dev, boot.part, boot.file);
 
-     if (strlen(boot.file)) {
-	  if (!strncmp(boot.file, "\\\\", 2))
-	       boot.file = "\\\\";
-	  else {
-	       char *p, *last;
-	       p = last = boot.file;
-	       while(*p) {
-		    if (*p == '\\')
-			 last = p;
-		    p++;
-	       }
-	       if (p)
-		    *(last) = 0;
-	       else
-		    boot.file = "";
-	       if (strlen(boot.file))
-		    strcat(boot.file, "\\");
-	  }
+     if (!conf_given) {
+         if (_machine == _MACH_chrp)
+             boot.file = "/etc/";
+         else if (strlen(boot.file)) {
+             if (!strncmp(boot.file, "\\\\", 2))
+                 boot.file = "\\\\";
+             else {
+                 char *p, *last;
+                 p = last = boot.file;
+                 while(*p) {
+                     if (*p == '\\')
+                         last = p;
+                     p++;
+                 }
+                 if (p)
+                     *(last) = 0;
+                 else
+                     boot.file = "";
+                 if (strlen(boot.file))
+                     strcat(boot.file, "\\");
+             }
+         }
+         strcpy(conf_path, boot.file);
+         strcat(conf_path, CONFIG_FILE_NAME);
+         boot.file = conf_path;
+         DEBUG_F("After path kludgeup: dev=%s, part=%d, file=%s\n",
+            boot.dev, boot.part, boot.file);
      }
-     DEBUG_F("After pmac path kludgeup: dev=%s, part=%d, file=%s\n",
-	     boot.dev, boot.part, boot.file);
 
-     useconf = load_config_file(boot.dev, boot.file, boot.part);
+     /*
+      * If we're doing a netboot, first look for one which matches our
+      * MAC address.
+      */
+     if (prom_get_devtype(boot.dev) == FILE_DEVICE_NET) {
+          prom_printf("Try to netboot\n");
+	  useconf = load_my_config_file(&boot);
+     }
+
+     if (!useconf)
+         useconf = load_config_file(&boot);
 
      prom_printf("Welcome to yaboot version " VERSION "\n");
      prom_printf("Enter \"help\" to get some basic usage information\n");
@@ -1521,16 +1739,18 @@ yaboot_main(void)
 	       if ((ptype != NULL) && (strcmp(ptype, "Apple_Bootstrap")))
 		    prom_printf("\nWARNING: Bootstrap partition type is wrong: \"%s\"\n"
 				"         type should be: \"Apple_Bootstrap\"\n\n", ptype);
+	       if (ptype)
+		    free(ptype);
 	  }
      }
 
      yaboot_text_ui();
-	
+
      prom_printf("Bye.\n");
      return 0;
 }
 
-/* 
+/*
  * Local variables:
  * c-file-style: "k&r"
  * c-basic-offset: 5
